/* Crusader serial reader; (C)2012 Jaroslav Fojtik, Licence GPL. */

// TODO Reagovat, pokud hlavice spadne do modu q.


#include <stdio.h>          /* Standard input/output definitions */
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <signal.h>
#include <fcntl.h>          /* File control definitions */
#include <termios.h>
#include <time.h>
#include <netinet/in.h>


#define WIFI_ONLY  4
#define WIFI_PREF  3
#define BALANCED   2
#define CRUS_PREF  1
#define CRUS_ONLY  0
static int OldStatus = -1;

#define MAX_LAST_PKT 32767

#ifndef _WIN32
  #define WSACleanup()		  /* nothing */
  #define WSAStartup(_x,_y)	0 /* return value 0 */
  #define closesocket(_s) close(_s)
  #define WSADATA int
#endif


#define _POSIX_SOURCE 1 /* POSIX compliant source */

/* 
baudrate settings are defined in <asm/termbits.h>, which is
   included by <termios.h> */
#define BAUDRATE B38400
int CRtsCts = CRTSCTS;
/* change this definition for the correct port */
const char *MODEMDEVICE = "/dev/ttyS0";

const char *CRUSADER_GW = "10.23.204.201";
const char *WIFI_GW = "10.23.202.254";
const char *CRUSADER_DEV = "eth2";
const char *WIFI_DEV = "wlan1";

const char *SHR_ROUTE = NULL;	// "default";

const char *ip_PATH = "/sbin/ip";
const char *rrdupdate_PATH = "/usr/sbin/rrdupdate";
const char *echo_PATH = "/usr/bin/echo";
const char *RRD_DATABASE = "/var/rrd/crusader.rrd";
const char *LOG_FILE = NULL;

const char *CRUS_STAT_FILE = "/tmp/crusader.txt";
FILE *LogFile = NULL;

char BASE_CHANNEL = '2';

int Verbosity = 0;
int LoadBalancing = 1;
#define FILTER_FACTOR	0.98

int F = -1;		/** Serial Crusader line. */


/* WAY 1 - WiFi line from peer to me */
int WiFiLastPkt = MAX_LAST_PKT;
float WiFiPass = 0;
unsigned R_WiFiSeq=0;	 /**< Sequence counter for WiFi. */


/* WAY 2 - Crusader line from peer to me */
int CrusLastPkt = 32767;
float CrusPass = 0;
unsigned R_CrusSeq=0;	 /**< Sequence counter for Crusader. */
int RSSI1;


/* WAY 3 - WiFi line from me to peer */
int R_WiFiLastPkt = MAX_LAST_PKT;
float R_WiFiPass = 0;
unsigned CrusSeq=0;	 /**< Sequence counter for Crusader. */


/* WAY 4 - Crusader line from me to peer */
int R_CrusLastPkt = MAX_LAST_PKT;
float R_CrusPass = 0;		/**< Reported crusader line pass from peer. */
unsigned WiFiSeq=0;		/**< Sequence counter for WiFi. */
int RSSI2;			/**< RSSI obtained from peer; -1 N/A. */


char R_Status = -1;


long SerialTimeout_us = 2500000;	/**< 2.5s serial line timeout. */
#define MINIMAL_TIMEOUT_us 200000	/**< 0.2s minimal timeout. */

unsigned MAX_RRD_RATE_MS = 1000;		/**< Maximal RRD update rate 1s */


struct termios old_tio;

/** Notifikacni struktura o stavu linky. */
typedef struct
{
  unsigned CRC;
  int RSSI;		 /**< Crusader RSSI */
  unsigned WiFiSeq;	 /**< Sequence counter for WiFi. */
  unsigned CrusSeq;      /**< Sequence counter for Crusader. */
  short int WiFiLastPkt; /**< Loop counter from last WiFi packet. */
  short int CrusLastPkt; /**< Loop counter from last Crusader packet. */
  float WiFiPass;
  float CrusPass;
  char  Status;		 /**< Line routing status. */
} CrusHdr;


char Line[256];
char RemoteLine[1500];

char packet[1500];

static volatile int SigTerm = 0;	///< Counter of sigterm signals. 2'nd sigterm calls abort.
static volatile int SigUsrCount = 0;	///< Counter of user signal.

int sock = -1;
struct sockaddr_in snd_commWiFi;
struct sockaddr_in snd_commCrus;
struct sockaddr_in recv_comm;
struct sockaddr_in sa_remote;
unsigned UDP_port = 2008;		///< UDP port for communication.


#define CHK(x) x==NULL?"NULL":x

const char *CrusaderStatusSTR(int status)
{
 switch(status)
 {
   case WIFI_ONLY: return "WIFI_ONLY";
   case WIFI_PREF: return "WIFI_PREF";
   case BALANCED:  return "BALANCED";
   case CRUS_PREF: return "CRUS_PREF";
   case CRUS_ONLY: return "CRUS_ONLY";
   case -1:	   return "Uninitialised";
 }
 return "Unknown";   
}

const char *FormatTime(void)
{
static char msg[64];

#if defined(_WIN32) && !defined(__MINGW32__)
    SYSTEMTIME  t;
      
    GetLocalTime(&t);		/* add time to message */  
    sprintf(msg, ("%04d/%02d/%02d %02d:%02d:%02d.%03d "),
      t.wYear, t.wMonth, t.wDay, t.wHour, t.wMinute, t.wSecond, t.wMilliseconds);
#else
 #ifdef __MINGW32__
    timeval t1;
    gettimeofday(&t1, NULL);
 #else
    struct timespec current_ntime;
    clock_gettime(CLOCK_REALTIME,&current_ntime);
 #endif
    time_t current_time;
    const struct tm *t;

    current_time=time((time_t *) NULL);
    t=localtime(&current_time);

    sprintf(msg, ("%04d/%02d/%02d %02d:%02d:%02d.%03d "),
      //mesice jsou cislovany od nuly, musime zobrazovat cislo o 1 vetsi
      t->tm_year+1900, t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec,
 #ifdef __MINGW32__
      (int)(t1.tv_usec/1000));
 #else
      (int)(current_ntime.tv_nsec/1000000));
 #endif
#endif

return msg;
}


int open_serial(const char *devname)
{
  int fd;
	/* Open modem device for reading and writing and not as controlling tty
	   because we don't want to get killed if linenoise sends CTRL-C. */
  fd = open(devname, O_RDWR | O_NOCTTY | O_NDELAY); 
  if(fd < 0) { perror(devname); }
  return fd;
}


int set_serial_line(int fd, struct termios *oldtio)
{
struct termios newtio;

  if(oldtio)
    tcgetattr(fd, oldtio); /* save current serial port settings */

  memset(&newtio, 0, sizeof(newtio)); /* clear struct for new port settings */

    /* BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
       CRTSCTS : output hardware flow control (only used if the cable has
	            all necessary lines. See sect. 7 of Serial-HOWTO)
       CS8     : 8n1 (8bit,no parity,1 stopbit)
       CLOCAL  : local connection, no modem contol
       CREAD   : enable receiving characters
       HUPCL   : Lower modem control lines after last process closes the device (hang up).  */
  newtio.c_cflag = BAUDRATE | CRtsCts | CS8 | CLOCAL | CREAD | HUPCL;
	 
    /* IGNPAR  : ignore bytes with parity errors
       ICRNL   : map CR to NL (otherwise a CR input on the other computer
                 will not terminate input)
                 otherwise make device raw (no other input processing) */
  newtio.c_iflag = IGNPAR | ICRNL;
         
    /* Raw output. */
  newtio.c_oflag = 0;
         
    /* ICANON  : enable canonical input
	  disable all echo functionality, and don't send signals to calling program */
  newtio.c_lflag = 0; 		// !!!!!! ICANON;
         
    /* initialize all control characters 
    default values can be found in /usr/include/termios.h, and are given
    in the comments, but we don't need them here */
  newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */ 
  newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
  newtio.c_cc[VERASE]   = 0;     /* del */
  newtio.c_cc[VKILL]    = 0;     /* @ */
  newtio.c_cc[VEOF]     = 4;     /* Ctrl-d */
  newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
  newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */
  newtio.c_cc[VSWTC]    = 0;     /* '\0' */
  newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */ 
  newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
  newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
  newtio.c_cc[VEOL]     = 0;     /* '\0' */
  newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
  newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
  newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
  newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
  newtio.c_cc[VEOL2]    = 0;     /* '\0' */

    /* now clean the modem line and activate the settings for the port. */
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd,TCSANOW, &newtio);

  return 0;
}


int restore_serial_line(int fd, struct termios *stored_tio)
{
	/* restore the old port settings */
  tcsetattr(fd, TCSANOW, stored_tio);
  return 0;
}



static long ReadInt(char *LineX)
{
char ch;
long n = 0;

  if(LineX==NULL) return 0;

  ch=*LineX++;
  while(isspace(ch) && ch!=0)
     {
     ch=*LineX++;
     }
     
  if(!isdigit(ch)) return -1;

  while(isdigit(ch) && ch!=0)
	{
	n=10*n+(ch-'0');
	ch=*LineX++;
	}

return n;
}


void ActualiseSeq(float *XPass, unsigned *R_Seq, unsigned SeqPkt, char Arrived)
{
int i;

  if(XPass==NULL || R_Seq==NULL) return;

  i = labs(SeqPkt - *R_Seq);

  if(i>=2)
  {
    *XPass = powf(FILTER_FACTOR,i-1) * *XPass;
    *R_Seq = SeqPkt - 1;
  }
  if(Arrived==0) return;
   
  *XPass = FILTER_FACTOR * *XPass + (1-FILTER_FACTOR)*Arrived;
  *R_Seq = SeqPkt;
}


/** Updare RRD database. */
void UpdateRSSI(int RSSI)
{
pid_t my_pid, parent_pid, child_pid;
int status;
char RSSI_S[20];
struct timespec tpCurr;
static struct timespec tpOld;
static int MinRSSI = -1;
static int MaxRSSI = -1;
char WriteCount = 0;

  if(rrdupdate_PATH==NULL || RRD_DATABASE==NULL) return;

	/* Update Min and Max to allow minimisation of RRD calls. */
  WriteCount++;
  if(RSSI>=0)
  {  
    if(RSSI>MaxRSSI) MaxRSSI=RSSI;
    if(RSSI<MinRSSI || MinRSSI<0) MinRSSI=RSSI;
  }

  clock_gettime(CLOCK_REALTIME, &tpCurr);
  if( labs((tpCurr.tv_sec*1000+tpCurr.tv_nsec/1000000) -
           (tpOld.tv_sec*1000+tpOld.tv_nsec/1000000)) < MAX_RRD_RATE_MS) return;

  if(RSSI>0) // Send Min and Max in odd and even round.
  {
    if(WriteCount & 1)
      {sprintf(RSSI_S,"N:%d",MinRSSI); MinRSSI = MaxRSSI;}
    else
      {sprintf(RSSI_S,"N:%d",MaxRSSI); MaxRSSI = MinRSSI;}
  }
  else	// RSSI lost, but Min and Max should be flushed.
  {
    if(MinRSSI>0)
      {sprintf(RSSI_S,"N:%d",MinRSSI); MinRSSI = -1;}
    else if(MaxRSSI>0)
      {sprintf(RSSI_S,"N:%d",MaxRSSI); MaxRSSI = -1;}
    else
      strcpy(RSSI_S,"N:U");		// Unknown RSSI
  }

  //printf("\n%s %s %s", rrdupdate_PATH, RRD_DATABASE, RSSI_S);

  if((child_pid = fork()) < 0)
    {
    if(LogFile)
      fprintf(LogFile,"Fork for RRUPDATE failure");
    rrdupdate_PATH=NULL;
    return;
    }

  if(child_pid == 0)	//a new-born process
  {
     // /usr/sbin/rrdupdate /var/rrd/rodgarssi.rrd N:$RODGASIGNAL
    execl(rrdupdate_PATH, rrdupdate_PATH, 
		RRD_DATABASE, RSSI_S, (char *)0);
    perror("Exec for RRUPDATE failed!");
    _exit(1);
    }

  if(Verbosity>=0) printf(" -");
  waitpid(child_pid,&status,0);
  if(Verbosity>=0) printf("* ");

  memcpy(&tpOld, &tpCurr, sizeof(tpOld));
}


/** Change routing weights. */
void SetRoutingStatus(int Status, const char *Reason)
{
pid_t my_pid, parent_pid, child_pid;
int status;

  if(Status==OldStatus) return;
  if(ip_PATH==NULL || SHR_ROUTE==NULL) return;

  if(Reason==NULL) Reason="";
  printf("\nRouting policy change %d->%d%s.",OldStatus,Status,Reason);
  if(LogFile!=stderr && LogFile!=NULL)
    fprintf(LogFile,"\n%s Routing policy change %d->%d%s.",FormatTime(),OldStatus,Status,Reason);

  my_pid = getpid();    parent_pid = getppid();

  if((child_pid = fork()) < 0)
    {
    if(LogFile)
      fprintf(LogFile,"\nFork failure");
    exit(1);
    }

  if(child_pid == 0)	//a new-born process
  {
    if(Status==CRUS_PREF)
      execl(ip_PATH, ip_PATH,
        "route","change",SHR_ROUTE,"scope","global",
          "nexthop","via",WIFI_GW,"dev",WIFI_DEV,"weight","1",
          "nexthop","via",CRUSADER_GW,"dev",CRUSADER_DEV,"weight","10",
	(char *) 0);

    if(Status==BALANCED)
      execl(ip_PATH, ip_PATH,
        "route","change",SHR_ROUTE,"scope","global",
          "nexthop","via",WIFI_GW,"dev",WIFI_DEV,"weight","10",
          "nexthop","via",CRUSADER_GW,"dev",CRUSADER_DEV,"weight","10",
	(char *) 0);

    if(Status==WIFI_PREF)
      execl(ip_PATH, ip_PATH,
        "route","change",SHR_ROUTE,"scope","global",
          "nexthop","via",WIFI_GW,"dev",WIFI_DEV,"weight","10",
          "nexthop","via",CRUSADER_GW,"dev",CRUSADER_DEV,"weight","1",
	(char *) 0);

    if(Status==WIFI_ONLY)
      execl(ip_PATH, ip_PATH,
        "route","change",SHR_ROUTE,"scope","global",
          "nexthop","via",WIFI_GW,"dev",WIFI_DEV,"weight","10",
	(char *)0);

    if(Status==CRUS_ONLY)
      execl(ip_PATH, ip_PATH,
        "route","change",SHR_ROUTE,"scope","global",
          "nexthop","via",CRUSADER_GW,"dev",CRUSADER_DEV,"weight","10",
	(char *)0);

    perror("Exec failed!");
    _exit(1);
    }

  printf("\n* ");
  waitpid(child_pid,&status,0);
  printf("Routing policy changed %d.",Status);
  OldStatus = Status;
}



/** This function prints help to the text console. */
void ShowHelp(void)
{
 printf("\nUsage: crus [switches]"
	"\nSwitches:"
	"\n  -?  --help   Display this help page.");
 printf("\n  --COMM_PORT dev     [%s] set crusader COM port", CHK(MODEMDEVICE));
 printf("\n  --UDP_PORT port     [%d] UDP communication port", UDP_port);
 printf("\n  --CRUSADER_GW GW    [%s] set crusader gateway", CHK(CRUSADER_GW));
 printf("\n  --WIFI_GW GW        [%s] set wifi gateway", CHK(WIFI_GW));
 printf("\n  --CRUSADER_DEV eth  [%s] set crusader ethernet device", CHK(CRUSADER_DEV));
 printf("\n  --WIFI_DEV eth      [%s] set crusader ethernet device", CHK(WIFI_DEV));

 printf("\n  --SHR_ROUTE route   [%s] set shared route", CHK(SHR_ROUTE));

 printf("\n  --ip_PATH ip_path   [%s] set path to 'ip' command", CHK(ip_PATH));
 printf("\n  --rrdupdate_PATH rr_path [%s] set path to 'rrdupdate' command", CHK(rrdupdate_PATH));

 printf("\n  --RRD_DATABASE db_path [%s] set path to rrd RSSI database", CHK(RRD_DATABASE));

 printf("\n  --BASE_CHANNEL val  [%c] set crusader base channel", BASE_CHANNEL);
 printf("\n  --CRUS_STAT_FILE file [%s] crusader status file", CHK(CRUS_STAT_FILE));
 printf("\n  --SERIAL_TIMEOUT val [%g] Timeout for serial line in [s].", SerialTimeout_us/1000000.0);
 printf("\n  --LOG_FILE file      [%s] file with logs", CHK(LOG_FILE));
 printf("\n  --MAX_RRD_RATE_MS val [%u] Maximal allowed rate for RRD database update.", MAX_RRD_RATE_MS);
 printf("\n  -s | -v             silent or verbose mode");

 printf("\n");
}


/** This function prints text console heading. */
void Heading(void)
{
static int printed = 0;

 if(printed) return;	//print heading only once
 printed = 1;

 printf("\nCrusader helper."
	"\nCopyright: (C) 2012 Jaroslav Fojtik, inspired by crusader100_read_rss\n");
 fflush(stdout);
}


/** Checks whether there is at least one additional argument. */
int CheckArg(int *p, int argc)
{
  if(p==NULL) return 0;
  if(++*p>=argc)
  {
    if(LogFile)
      fprintf(LogFile,"\nMissing argument!");
    return 0;  
  }

return 1;  
}


/** This procedure handles parameters and the filenames.
 * @param[in] DbHandler		Databaze connetion parameters.
 * @return	 		0 - parameters read; 1 - exit program */
int ProcessCommandLine(int argc, char *argv[])
{
  char *StrName = NULL;
  const char *name;
  int p;
  char firstch;

  for(p=1; p<argc; p++)
    {
    if(StrName) {free(StrName); StrName=NULL;}
    StrName = strdup(argv[p]);
    if(StrName=="") continue;

    firstch = StrName[0];
    if (StrName[0] == '/') StrName[0] = '-';
    name = StrName;

    if (name[0] == '-')
      {
      if(!strcmp(name, "--BASE_CHANNEL"))
         {
         if(!CheckArg(&p,argc)) continue;
	 BASE_CHANNEL = *argv[p];
	 if(!isdigit(BASE_CHANNEL))
	   {
	   if(LogFile)
	     fprintf(LogFile,"\nWrong BASE_CHANNEL %c, changed to 1",BASE_CHANNEL);
	   BASE_CHANNEL = '1';
           }
	 continue;
         }

      if(!strcmp(name, "--COMM_PORT"))
         {
         if(!CheckArg(&p,argc)) continue;
	 MODEMDEVICE = argv[p];
	 continue;
         }

      if(!strcmp(name, "--CRUS_STAT_FILE"))
         {
         if(!CheckArg(&p,argc)) continue;
	 CRUS_STAT_FILE = argv[p];
	 continue;
         }

      if(!strcmp(name, "--CRUSADER_DEV"))
         {
         if(!CheckArg(&p,argc)) continue;
	 CRUSADER_DEV = argv[p];
	 continue;
         }

      if(!strcmp(name, "--CRTSCTS"))
         {
	 CRtsCts = CRTSCTS;
	 continue;
         }

      if(!strcmp(name, "--NO_CRTSCTS"))
         {
	 CRtsCts = 0;
	 continue;
         }

      if(!strcmp(name, "--CRUSADER_GW"))
         {
         if(!CheckArg(&p,argc)) continue;
	 CRUSADER_GW = argv[p];
	 continue;
         }

      if(!strcmp(name, "--help") || !strcmp(name, "-?") || !strcmp(name, "-help"))
	   {
	   Heading();
	   ShowHelp();
	   return(1);
	   }

      if(!strcmp(name, "--ip_PATH") || !strcmp(name, "--IP_PATH"))
        {
        if(!CheckArg(&p,argc)) continue;
	ip_PATH = argv[p];
	continue;
        }

      if(!strcmp(name, "--LOG_FILE"))
        {
        if(!CheckArg(&p,argc)) continue;
	LOG_FILE = argv[p];
	continue;
        }

      if(!strcmp(name, "--MAX_RRD_RATE_MS"))
        {
        if(!CheckArg(&p,argc)) continue;	 
	MAX_RRD_RATE_MS = atoi(argv[p]);
	continue;
        }

      if(!strcmp(name, "--MAX_RRD_RATE"))
         {
         if(!CheckArg(&p,argc)) continue;	 
	 MAX_RRD_RATE_MS = 1000*atof(argv[p]);
	 continue;
         }

      if(!strcmp(name, "--NO_ip") || !strcmp(name, "--NO_IP"))
         {
	 ip_PATH = NULL;
	 continue;
         }

      if(!strcmp(name, "--NO_rrdupdate") || !strcmp(name, "--NO_RRUPDATE"))
         {
	 rrdupdate_PATH = NULL;
	 continue;
         }

      if(!strcmp(name, "--NO_SHARED_ROUTE"))
         {
	 LoadBalancing = 0;
	 continue;
         }

      if(!strcmp(name, "--SERIAL_TIMEOUT"))
         {
         if(!CheckArg(&p,argc)) continue;	 
	 SerialTimeout_us = 1000000*atof(argv[p]);
	 if(SerialTimeout_us<MINIMAL_TIMEOUT_us)
	   {
	   if(LogFile)
	     fprintf(LogFile,"\nBad serial timeout \"%g\".", SerialTimeout_us/1000000.0);
	   SerialTimeout_us=MINIMAL_TIMEOUT_us;
	   }
	 continue;
         }

      if(!strcmp(name, "--SHARED_ROUTE"))
         {
	 LoadBalancing = 1;
	 continue;
         }

      if(!strcmp(name, "--WIFI_DEV"))
         {
         if(!CheckArg(&p,argc)) continue;
	 WIFI_DEV = argv[p];
	 continue;
         }

      if(!strcmp(name, "--rrdupdate_PATH") || !strcmp(name, "--RRUPDATE_PATH"))
         {
         if(!CheckArg(&p,argc)) continue;
	 rrdupdate_PATH = argv[p];
	 continue;
         }

      if(!strcmp(name, "--RRD_DATABASE"))
         {
         if(!CheckArg(&p,argc)) continue;
	 RRD_DATABASE = argv[p];
	 continue;
         }

      if(!strcmp(name, "--SHR_ROUTE"))
         {
         if(!CheckArg(&p,argc)) continue;
	 SHR_ROUTE = argv[p];
	 continue;
         }

      if(!strcmp(name, "--UDP_PORT"))
         {
         if(!CheckArg(&p,argc)) continue;
	 UDP_port = atol(argv[p]);
	 continue;
         }

      if(!strcmp(name, "--WIFI_GW"))
         {
         if(!CheckArg(&p,argc)) continue;
	 WIFI_GW = argv[p];
	 continue;
         }

      /*if(!strcmp(name, "-l"))
         {
         if(!CheckArg(p,argc)) continue;
         TgtLog.Level = (MSGCLASS)atol(argv[p]);
	 continue;
         }*/

       if(!strcmp(name, "-s"))
         {
         Verbosity--;         
	 continue;
         }

       if(!strcmp(name, "-v"))
         {
         Verbosity++;
	 continue;
         }

      StrName[0] = firstch;
      if(firstch=='-') goto UnknownArg;
      }

UnknownArg:
    Heading();
    if(LogFile)
      fprintf(LogFile,"\nUnknown argument \"%s\"!",StrName);
    SigTerm=1;
  }

  if(StrName) {free(StrName); StrName=NULL;}
return(0);
}


/*****************************************************/
/** This is a main signal handler. */
void sig_int(int signo)
{
// printf("Signal received %d!"); fflush(stdout);

 switch(signo)
    {
#ifdef SIGUSR1
      case SIGUSR1: SigUsrCount++;       // SIGUSR1 breaks a delay.
		    if(LogFile==NULL)
		      {
		      if(LOG_FILE!=NULL)
			{
			LogFile = fopen(LOG_FILE,"ab");
		        }
		      }
		    else
		      {
		      if(LogFile!=stderr)
		        {
		        fprintf(LogFile,"\n%s LogFile detached because of SIGUSR1 received.", FormatTime());
		        fclose(LogFile); //SetEvent(MainAlertEvent);
		        LogFile = NULL;
		        }
		      }
		    break;
#endif

#ifdef SIGUSR2
      case SIGUSR2: 
		    if(F>=0)
		      {
		      if(LogFile)
		        fprintf(LogFile,"\n%s Serial line detached because of SIGUSR2 received.", FormatTime());
		      // restore_serial_line(F, &old_tio);
		      close(F);
		      F =- 1;
		      }
		    else
		      {
		      if(MODEMDEVICE!=NULL && *MODEMDEVICE!=0)
		        {
			F = open_serial(MODEMDEVICE);
			if(F>=0)
			  {
			  if(LogFile)
	                    fprintf(LogFile,"\n%s Serial line reopened because of SIGUSR2 received.", FormatTime());
			  set_serial_line(F,NULL);
			  write(F, "Z\r", 2);		// Check for base.
			  }
			}
		      }
		    break;
#endif

      case SIGINT:
      case SIGTERM: if(SigTerm++>=2)
		      {
		      if(LogFile)
		        fprintf(LogFile,"\n%s 2nd SIGTERM has been received.",FormatTime());
		      abort();
		      }
		    else
		      {
		      if(LogFile)
		        fprintf(stderr,"\n%s SIGTERM has been received - scheduling finish.",FormatTime());
		      }
		    //SetEvent(MainAlertEvent);
		    break;		    
      case SIGSEGV: {
		    if(LogFile)
		      fprintf(stderr,"\n%s Access violation, please report a bug.",FormatTime());
                    break;
		    }
    }
}


static void HookSignals(void)
{
#ifdef SIGUSR1
  signal(SIGUSR1,sig_int);
#endif
#ifdef SIGUSR2
  signal(SIGUSR2,sig_int);
#endif
  signal(SIGINT,sig_int);
  signal(SIGTERM,sig_int);
  signal(SIGSEGV,sig_int);
}


/** Read one line crusader report from the serial line. Also try t read
  remote line when available. */
int ReadCrudaderLine(int F)
{
char c;
int i, size;
int precteno;
int Timeout = SerialTimeout_us;

  while(Timeout>0 && SigTerm<=0)
  {
    size = sizeof(sa_remote);

    do {
       i = recvfrom(sock, packet, sizeof(packet), 0, (struct sockaddr *)&sa_remote, &size);
       if(i>0)
         {
	   if(snd_commWiFi.sin_addr.s_addr==sa_remote.sin_addr.s_addr)
	     {
	     WiFiLastPkt = 0;
	     R_WiFiLastPkt = ((CrusHdr*)(packet))->WiFiLastPkt;
	     R_CrusLastPkt = ((CrusHdr*)(packet))->CrusLastPkt;
	     R_Status = ((CrusHdr*)(packet))->Status;
	     R_WiFiPass = ((CrusHdr*)(packet))->WiFiPass;
	     R_CrusPass = ((CrusHdr*)(packet))->CrusPass;
	     ActualiseSeq(&WiFiPass, &R_WiFiSeq, ((CrusHdr*)(packet))->WiFiSeq, 1);
             ActualiseSeq(&CrusPass, &R_CrusSeq, ((CrusHdr*)(packet))->CrusSeq, 0);
	     if(i>sizeof(CrusHdr))
	       memcpy(RemoteLine,packet+sizeof(CrusHdr),i-sizeof(CrusHdr));	     
	     }
           else if(snd_commCrus.sin_addr.s_addr==sa_remote.sin_addr.s_addr)
	     {
	     CrusLastPkt = 0;
	     R_WiFiLastPkt = ((CrusHdr*)(packet))->WiFiLastPkt;
	     R_CrusLastPkt = ((CrusHdr*)(packet))->CrusLastPkt;
	     R_Status = ((CrusHdr*)(packet))->Status;
	     R_WiFiPass = ((CrusHdr*)(packet))->WiFiPass;
	     R_CrusPass = ((CrusHdr*)(packet))->CrusPass;
             ActualiseSeq(&CrusPass, &R_CrusSeq, ((CrusHdr*)(packet))->CrusSeq, 1);
	     ActualiseSeq(&WiFiPass, &R_WiFiSeq, ((CrusHdr*)(packet))->WiFiSeq, 0);
	     if(i>sizeof(CrusHdr))
	       memcpy(RemoteLine,packet+sizeof(CrusHdr),i-sizeof(CrusHdr));	     
	     }
           else
             {
	     printf("\nBogus UDP received from %s:%d.",inet_ntoa(sa_remote.sin_addr),ntohs(sa_remote.sin_port));
	     memset(RemoteLine, 0, sizeof(RemoteLine));
	     }
         }
       } while(i>0);

    //if(Timeout<10000) printf(" &&& %d &&& ", Timeout);
    c = 0;
    precteno = read(F, &c, 1);
    if(precteno<=0) 
      {usleep(4000);Timeout-=4000;continue;}
    if(Verbosity>=0) putchar(c);
    if(c=='!')
    {
      do {
         precteno = read(F, &c, 1);
         if(precteno<=0) {usleep(500);Timeout-=500;}
         } while(precteno<=0 && Timeout>=0);
      putchar(c);
      if(c!='O') continue;

      do {
         precteno = read(F, &c, 1);
         if(precteno<=0) {usleep(500);Timeout-=500;}
         } while(precteno<=0 && Timeout>=0);
      putchar(c);
      if(c!='K') continue;

      return -2;
    }

    if(c!='+') continue;

    do {
       precteno = read(F, &c, 1);
       if(precteno<=0) {usleep(500);Timeout-=500;}
       } while(precteno<=0 && Timeout>=0);

    if(c!=' ')
      {
      if(c!='+') continue;

      do {
         precteno = read(F, &c, 1);
         if(precteno<=0) {usleep(500);Timeout-=500;}
         } while(precteno<=0 && Timeout>=0);
      }
    if(c!=' ') continue;

    if(Verbosity>=0) printf(">>>");

    memset(Line,0,sizeof(Line));
    for(i=0; i<sizeof(Line)-1; i++)
      {
      do {
         precteno = read(F, &c, 1);
         if(precteno<=0) {usleep(900);Timeout-=900;}
         } while((precteno<=0 || c==0) && Timeout>=0);
      Line[i] = c;
      if(Line[i]=='\n' || Timeout<0) 
        {Line[i]=0;break;}
      }
    break;
  }

 //printf(" &&& %d &&& ", Timeout);
 return (Timeout>0)?0:-1;
}


/** Send notification packet to counterpart. */
void NotifyUDP(int sock, struct sockaddr_in *snd_comm, const char *MessageData, int MyRSSI)
{
int len, ret;
CrusHdr H;

  if(sock<0 || snd_comm->sin_port==0) return;

  len = strlen(MessageData);
  if(len>1024) len=1024;

  H.CRC = 0;
  H.RSSI = MyRSSI;
  H.WiFiLastPkt = WiFiLastPkt;
  H.CrusLastPkt = CrusLastPkt;
  H.Status = OldStatus;
  H.WiFiSeq = WiFiSeq;
  H.CrusSeq = CrusSeq;
  H.WiFiPass = WiFiPass;
  H.CrusPass = CrusPass;

  memcpy(packet,&H,sizeof(CrusHdr));
  memcpy(packet+sizeof(CrusHdr),MessageData,len+1);

  ret = sendto(sock, (const char *)packet, len+1+sizeof(CrusHdr), 0, (struct sockaddr *)snd_comm, sizeof(struct sockaddr_in));
}


void BlockingMode(int s, u_long arg)
{
#if defined(_WIN32) || defined(_WIN64)
  ioctlsocket(socketUdpId, FIONBIO, &arg);
#else
int x;
  x = fcntl(s,F_GETFL,0);
  if(arg)
    fcntl(s,F_SETFL, x | O_NONBLOCK);
  else
    fcntl(s,F_SETFL, x & ~O_NONBLOCK);
#endif
}


volatile int UDP_Instances = 0;


int StartupUDP(void)
{
WSADATA wsaData;
  if(UDP_Instances>0) return 1;
  if(WSAStartup(MAKEWORD(2,0), &wsaData)) return -1;
  UDP_Instances++;
  return 1;
}

int CloseUDP(void)
{
  UDP_Instances--;
  if(UDP_Instances>0) return 1;
  WSACleanup();
}


/** Inintialize UDP subsystem. */
void SetupUDP(void)
{
unsigned long arg;
  if(UDP_port<=0) return;

  if(StartupUDP()<=0)
    {
    if(LogFile)
      fprintf(LogFile,"\nUDP initialisation failure.");        
    return;
    }

	/* Create socket from which to read */
  sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(sock < 0)   
    {
    if(LogFile)
      fprintf(LogFile,"\nOpening UDP datagram socket failed.");
    UDP_port = 0;
    CloseUDP();
    return;
    }

  memset((char *)&recv_comm, 0, sizeof(recv_comm));
  recv_comm.sin_family = AF_INET;  
  recv_comm.sin_addr.s_addr = htonl(INADDR_ANY); //inet_addr("127.0.0.1");  //htonl(INADDR_ANY);  
  recv_comm.sin_port = htons(UDP_port);

  if(bind(sock, (struct sockaddr *)&recv_comm, sizeof(recv_comm))) 
    {
    closesocket(sock);
    sock = -1;
    CloseUDP();
    if(LogFile)
      fprintf(LogFile,"\nBinding datagram to socket %d failed.",UDP_port);
    return;
    }

  BlockingMode(sock, 1);  /* Set non-blocking receiving mode. */

  memset((char *)&snd_commWiFi, 0, sizeof(snd_commWiFi));
  snd_commWiFi.sin_family = AF_INET;
  if(WIFI_GW!=NULL)
    {
    snd_commWiFi.sin_addr.s_addr = inet_addr(WIFI_GW);
    snd_commWiFi.sin_port = htons(UDP_port);
    }

  memset((char *)&snd_commCrus, 0, sizeof(snd_commCrus));
  snd_commCrus.sin_family = AF_INET;
  if(CRUSADER_GW!=NULL)
    {
    snd_commCrus.sin_addr.s_addr = inet_addr(CRUSADER_GW);
    snd_commCrus.sin_port = htons(UDP_port);
    }
}


/** Main worker. */
int main(int argc, char *argv[])
{
FILE *F2;
char c;
int i;
int RSSI;
int SerialLineStatus = 1;

  LogFile = stderr;
  if(ProcessCommandLine(argc,argv)!=0)
    return 1;
  if(LOG_FILE!=NULL)
    {
      LogFile = fopen(LOG_FILE,"ab");
      if(LogFile==NULL)
        LogFile = stderr;
      else
        fprintf(LogFile,"\n%s Crusader governor started.",FormatTime());
    }

  F = open_serial(MODEMDEVICE);
  if(F<0)
  {
    if(LogFile)
      fprintf(LogFile, "\n%s Cannot open serial line (%s).", FormatTime(), MODEMDEVICE);
    return -1;
  }

  HookSignals();		/* Hook signals like CTRL+C. */
  SetupUDP();			/* Initialise UDP subsystem. */

  if(F>=0)
    {
    set_serial_line(F, &old_tio);
    write(F, "Z\r", 2);		// Check for base.
    }

  memset(RemoteLine,0,sizeof(RemoteLine));
  i=0;

  RSSI1 = 0;
  while(SigTerm<=0)
    {
    RSSI2 = -1;
    if(WiFiLastPkt<MAX_LAST_PKT) WiFiLastPkt++;
    if(CrusLastPkt<MAX_LAST_PKT) CrusLastPkt++;
    if(CrusLastPkt>100 && WiFiLastPkt>100) 	/* Crusader connection lost. */
      {
      if(WiFiLastPkt<MAX_LAST_PKT && CrusLastPkt<MAX_LAST_PKT && R_Status>=0)
        {  
        if(LogFile)
          fprintf(LogFile,"\n%s Twin Crusader governor hanged.",FormatTime());
        }
      memset(RemoteLine, 0, sizeof(RemoteLine));
      R_WiFiLastPkt = R_CrusLastPkt = -1;
      R_Status = -1;
      }

    i = ReadCrudaderLine(F);
    if(RemoteLine[0]!=0)
        RSSI2 = ReadInt(RemoteLine);

    F2 = fopen(CRUS_STAT_FILE,"wb");
    if(F2)
      {
      fprintf(F2,"Local status: %s\n%s\n",
		CrusaderStatusSTR(OldStatus), Line);

      if(RemoteLine[0]!=0)
	{
	fprintf(F2,"\nRemote response: %s\n%s:%d %s",
	        CrusaderStatusSTR(R_Status),
		inet_ntoa(sa_remote.sin_addr),
		ntohs(sa_remote.sin_port),
		RemoteLine);

        fprintf(F2,"\nWay 1 peer->me  R_WiFi ping: %d, R_WiFi loss %1.1f%%.",
			WiFiLastPkt, 100*(1-WiFiPass));
        fprintf(F2,"\nWay 2 peer->me  R_Crusader ping: %d, Crusader loss %1.1f%%, RSSI %d.",
			CrusLastPkt, 100*(1-CrusPass), RSSI1);
        fprintf(F2,"\nWay 3 me->peer  R_WiFi ping: %d, R_WiFi loss %1.1f%%.",
			R_WiFiLastPkt, 100*(1-R_WiFiPass));
        fprintf(F2,"\nWay 4 me->peer  R_Crusader ping: %d, Crusader loss %1.1f%%, RSSI %d.",
			R_CrusLastPkt, 100*(1-R_CrusPass), RSSI2);
        }
      else
        fprintf(F2,"WiFi ping: %d, Crusader ping %d.", WiFiLastPkt, CrusLastPkt);

      fclose(F2);
      }

    if(i>=0)
      {
      SerialLineStatus = 1;      
      RSSI1 = ReadInt(Line);

      if(Verbosity>0)
	printf(" RSSI %d, RMT_RSSI %d, CrusLine: %d, WiFi line %d || %s\r", 
		RSSI1, RSSI2, CrusLastPkt, WiFiLastPkt, Line);
      else if(Verbosity>=0)
        printf(" RSSI %d, RMT_RSSI %d \r", RSSI1, RSSI2);

      RSSI = RSSI1;
      if(RSSI2>=0)		// Calculate minimum from both RSSIs.
      {
        if(RSSI2<RSSI1) RSSI = RSSI2;
      }

      if((R_CrusLastPkt>55 || CrusLastPkt>55) && 
	  WiFiLastPkt<40 && labs(CrusLastPkt-WiFiLastPkt)>20)
	{  // UDP packets through Crusader has been hanged.
	  SetRoutingStatus(WIFI_ONLY," UDP through Crusader hanged");
	}
      else if(WiFiLastPkt>55 && CrusLastPkt<40 && labs(CrusLastPkt-WiFiLastPkt)>20)
	{  // UDP packets through WiFi has been hanged.
	  SetRoutingStatus(CRUS_ONLY," UDP through WiFi hanged");
	}
      else
	{
        if(LoadBalancing)
          {
          if(R_CrusPass<0.98) SetRoutingStatus(WIFI_ONLY," Crusader packetloss");
          else if(RSSI<120) SetRoutingStatus(WIFI_ONLY," Bad reported RSSI");
          else if(RSSI<150) SetRoutingStatus(WIFI_PREF," Medicore reported RSSI");
          else if(RSSI<200) SetRoutingStatus(BALANCED," Good reported RSSI");
          else SetRoutingStatus(CRUS_PREF," Excellent reported RSSI");
          }
        else
	  {	/* When shared route is not available, use hysteresis. */
	  if(R_CrusPass<0.98) SetRoutingStatus(WIFI_ONLY," Crusader packetloss");
	  else if(RSSI<140) SetRoutingStatus(WIFI_ONLY," Bad reported RSSI");
          else if(RSSI>170) SetRoutingStatus(CRUS_ONLY," Good reported RSSI");
          }
        }
      }
    else	/* i>=0 */
      {
      strcpy(Line,"Crusader serial lost");
      if(i==-2)
        {
	printf("\nBase station echo");
        write(F,"#", 1);             	// Exit from base station.
        write(F,&BASE_CHANNEL,1);       // Switch to a given crusader.
	write(F,"\r", 1);	
        }
      else
        {
        if(Verbosity>0)
          printf("\nCrusader serial lost, R_RSSI %d, CrusLine: %d, WiFi line %d",
		 RSSI2, CrusLastPkt, WiFiLastPkt);
	else if(RSSI1!=-1)
	  printf("\nCrusader serial lost");

         
	if(SerialLineStatus==10 && LogFile)
          fprintf(LogFile, "\n%s Serial line lost.", FormatTime());

	SerialLineStatus++;
        if(MODEMDEVICE!=NULL)
        {
	  if(SerialLineStatus % (2*60*60) == 0)
	    {
	    if(F>=0)
	      {
              // restore_serial_line(F, &old_tio);
	      close(F);		/* this emits HUP because of HUPCL */
	      F =- 1;
              }
	    }
	  if(SerialLineStatus % (2*60*60) == 4 && F<0)
	    {
	    F = open_serial(MODEMDEVICE);
	    if(F>=0)
	      {
	      set_serial_line(F,NULL);
	      write(F, "Z\r", 2);		// Check for base.
	      if(LogFile)
	        fprintf(LogFile, "\n%s Serial line reset", FormatTime());
	      }
	    else
	      {
	      if(LogFile)
		fprintf(LogFile, "\n%s Cannot open serial line (%s)",FormatTime(), MODEMDEVICE);
	      }
            }
	  }
        }
      RSSI1 = -1;

	/* Try to rely on UDP packets passing through both lines. */
      if((R_CrusLastPkt>55 || CrusLastPkt>55) && 
	  WiFiLastPkt<40 && labs(CrusLastPkt-WiFiLastPkt)>20)
	{  // UDP packets through Crusader has been hanged.
	  SetRoutingStatus(WIFI_ONLY," UDP through Crusader hanged");
	}
      else if(WiFiLastPkt>55 && CrusLastPkt<40 && labs(CrusLastPkt-WiFiLastPkt)>20)
	{  // UDP packets through WiFi has been hanged.
	  SetRoutingStatus(CRUS_ONLY," UDP through WiFi hanged");
	}
      else
	{
	if(CrusLastPkt<40 && WiFiLastPkt<40)
	  {
	  if(RSSI2>=0)
            {
	    if(LoadBalancing)
              {
              if(R_CrusPass<0.98) SetRoutingStatus(WIFI_ONLY," Crusader packetloss");
              else if(RSSI2<120) SetRoutingStatus(WIFI_ONLY," Bad reported R_RSSI");
              else if(RSSI2<150) SetRoutingStatus(WIFI_PREF," Medicore reported R_RSSI");
              else if(RSSI2<200) SetRoutingStatus(BALANCED," Good reported R_RSSI");
              else SetRoutingStatus(CRUS_PREF," Excellent reported R_RSSI");
              }
            else
	      {	/* When shared route is not available, use hysteresis. */
	      if(R_CrusPass<0.98) SetRoutingStatus(WIFI_ONLY," Crusader packetloss");
	      else if(RSSI2<140) SetRoutingStatus(WIFI_ONLY," Bad reported R_RSSI");
              else if(RSSI2>170) SetRoutingStatus(CRUS_ONLY," Good reported R_RSSI");
              }
            }
	    else
	      {
	      if(R_CrusPass<0.98) SetRoutingStatus(WIFI_ONLY," Crusader packetloss");
	      else
		{
		if(LoadBalancing)
		  SetRoutingStatus(BALANCED," Packets are passing, RSSI lost");
		else
		  SetRoutingStatus(CRUS_ONLY," Packets are passing, RSSI lost");
		}
	      }
	  }
        else
	  {
	  SetRoutingStatus(WIFI_ONLY," Serial & ping lost");
	  } 
	}
      }

    UpdateRSSI(RSSI1);
    NotifyUDP(sock, &snd_commWiFi, Line, RSSI1);
    NotifyUDP(sock, &snd_commCrus, Line, RSSI1); 
    WiFiSeq++;
    CrusSeq++;
    }

	/* Exit program. */    
  SetRoutingStatus(WIFI_ONLY," Program exit");

  if(F>=0)
    {
    restore_serial_line(F, &old_tio);
    close(F);
    F =- 1;
    }

  F2 = fopen(CRUS_STAT_FILE,"ab");
  if(F2)
    {
    fprintf(F2,"\nCrusader governor is not active.");
    fprintf(F2,"\nLast response:\n%s",Line);
    if(RemoteLine[0]!=0)
      fprintf(F2,"\nLast remote response:\n%s",RemoteLine);
    fprintf(F2,"\n");
    fclose(F2);
    }

  if(sock>=0)
  {
    closesocket(sock);
    sock = -1;
    CloseUDP();
  }
  
  if(LogFile!=stderr && LogFile!=NULL)
  {
     fclose(LogFile);
     LogFile = NULL;
  }

return 0;
}
